import 'dart:async';
import 'dart:convert';
import 'package:http/http.dart' as http;
import 'types.dart';

class RequireAccessToken {
  const RequireAccessToken();
  String toString() =>
      'You need to fetch a valid access token to do API calls.';
}

/// A fetched access token.
class AccessToken {
  /// The access token itself.
  String get accessToken => _accessToken;
  String _accessToken = '';

  /// When will the acccess token expire.
  int get expiresIn => _expiresIn;
  int _expiresIn = 0;

  /// Type of the access token.
  String get tokenType => _tokenType;
  String _tokenType = '';

  AccessToken(Map<String, dynamic> jsonObject) {
    this._accessToken = jsonObject['access_token'];
    this._tokenType = jsonObject['token_type'];
    this._expiresIn = jsonObject['expires_in'];

    if (this._accessToken.length == 0) {
      throw new Exception('Invalid access token');
    }
  }

  String toString() {
    return '<Access Token: ${this.accessToken} , Token Type: ${this.tokenType}, Expires In: ${this.expiresIn.toString()}>';
  }
}

/// The territory where KKBOX provides services.
enum Territory { taiwan, hongkong, singapore, malaysia, japan }

_territoryToString(Territory territory) {
  switch (territory) {
    case Territory.taiwan:
      return 'TW';
    case Territory.hongkong:
      return 'HK';
    case Territory.singapore:
      return 'SG';
    case Territory.malaysia:
      return 'MY';
    case Territory.japan:
      return 'JP';
  }
}

/// The types used in doing search.
enum SearchType { artist, album, track, playlist }

/// Helps to interact with KKBOX's Open API.
class KKBOXOpenAPI {
  static const String _oauthTokenUrl = 'https://account.kkbox.com/oauth2/token';
  static const String _apiBase = 'https://api.kkbox.com/v1.1/';

  String _clientID = '';
  String _clientSecret = '';
  AccessToken _accessToken = null;

  /// The constructor.
  KKBOXOpenAPI(String clientID, String clientSecret) {
    this._clientID = clientID;
    this._clientSecret = clientSecret;
  }

  /// If a valid access token is fetched.
  bool get isLoggedIn => this._accessToken != null;

  /// Removes the existing access token.
  void logout() {
    this._accessToken = null;
  }

  /// Fetches a new access token.
  Future<AccessToken> fetchAccessTokenByClientCredential() async {
    base64(String data) {
      var bytes = UTF8.encode(data);
      return BASE64.encode(bytes);
    }

    final credential = base64(this._clientID + ':' + this._clientSecret);
    final client = new http.Client();
    final response = await client.post(KKBOXOpenAPI._oauthTokenUrl, body: {
      'grant_type': 'client_credentials'
    }, headers: {
      'Authorization': 'Basic ' + credential,
      'Content-type': 'application/x-www-form-urlencoded'
    }).whenComplete(client.close);

    if (response.statusCode != 200) {
      throw new Exception(response.statusCode);
    }
    final jsonObject = JSON.decode(response.body);
    this._accessToken = new AccessToken(jsonObject);
    return this._accessToken;
  }

  _get(url) async {
    final escapedUrl = Uri.encodeFull(url);
    if (this._accessToken == null) {
      throw new Exception(
          'You need to fetch a valid access token to do API calls');
    }
    final client = new http.Client();
    final response = await client.get(escapedUrl, headers: {
      'Authorization': 'Bearer ' + this._accessToken.accessToken
    }).whenComplete(client.close);
    if (response.statusCode != 200) {
      throw new Exception(response.statusCode);
    }
    var jsonObject = JSON.decode(response.body);
    return jsonObject;
  }

  /// Fetches metadata of a track by giving a [trackID].
  ///
  /// See `(https://docs-en.kkbox.codes/v1.1/reference#tracks-track_id`.
  @RequireAccessToken()
  Future<TrackInfo> fetchTrack(String trackID,
      [Territory territory = Territory.taiwan]) async {
    final url = KKBOXOpenAPI._apiBase +
        'tracks/$trackID?territory=${_territoryToString(territory)}';
    final results = await this._get(url);
    return new TrackInfo(results);
  }

  /// Fetches metadata of an album by giving an [albumID]
  ///
  /// See `https://docs-en.kkbox.codes/v1.1/reference#albums-album_id`.
  @RequireAccessToken()
  Future<AlbumInfo> fetchAlbum(String albumID,
      [Territory territory = Territory.taiwan]) async {
    final url = KKBOXOpenAPI._apiBase +
        'albums/$albumID?territory=${_territoryToString(territory)}';
    final results = await this._get(url);
    return new AlbumInfo(results);
  }

  /// Fetches tracks contained in an album by giving an [albumID].
  ///
  /// See `https://docs-en.kkbox.codes/v1.1/reference#albums-album_id-tracks`.
  @RequireAccessToken()
  Future<TrackList> fetchTracksInAlbum(String albumID,
      [Territory territory = Territory.taiwan]) async {
    final url = KKBOXOpenAPI._apiBase +
        'albums/$albumID/tracks?territory=${_territoryToString(territory)}';
    final results = await this._get(url);
    return new TrackList(results);
  }

  /// Fetches metadata of an artist by giving [artistID].
  ///
  /// See `https://docs-en.kkbox.codes/v1.1/reference#artists-artist_id`.
  @RequireAccessToken()
  Future<ArtistInfo> fetchArtist(String artistID,
      [Territory territory = Territory.taiwan]) async {
    final url = KKBOXOpenAPI._apiBase +
        'artists/$artistID?territory=${_territoryToString(territory)}';
    final results = await this._get(url);
    return new ArtistInfo(results);
  }

  /// Fetches albums belong to an artist by giving an [artistID].
  ///
  /// See `https://docs-en.kkbox.codes/v1.1/reference#artists-artist_id-albums`.
  @RequireAccessToken()
  Future<AlbumList> fetchAlbumsOfArtist(String artistID,
      [Territory territory = Territory.taiwan,
      int offset = 0,
      int limit = 200]) async {
    final url = KKBOXOpenAPI._apiBase +
        'artists/$artistID/albums?territory=${_territoryToString(territory)}&offset=$offset&limit=$limit';
    final results = await this._get(url);
    return new AlbumList(results);
  }

  /// Fetches top tracks of an artist by giving an [artistID],
  ///
  /// See `https://docs-en.kkbox.codes/v1.1/reference#artists-artist_id-toptracks`.
  @RequireAccessToken()
  Future<TrackList> fetchTopTracksOfArtist(String artistID,
      [Territory territory = Territory.taiwan,
      int offset = 0,
      int limit = 200]) async {
    final url = KKBOXOpenAPI._apiBase +
        'artists/$artistID/top-tracks?territory=${_territoryToString(territory)}&offset=$offset&limit=$limit';
    final results = await this._get(url);
    return new TrackList(results);
  }

  /// Fetches related artists of an artist by giving an [artistID],
  ///
  /// See `https://docs-en.kkbox.codes/v1.1/reference#artists-artist_id-relatedartists`.
  @RequireAccessToken()
  Future<ArtistList> fetchRelatedArtistsOfArtist(String artistID,
      [Territory territory = Territory.taiwan,
      int offset = 0,
      int limit = 20]) async {
    final url = KKBOXOpenAPI._apiBase +
        'artists/$artistID/related-artists?territory=${_territoryToString(territory)}&offset=$offset&limit=$limit';
    final results = await this._get(url);
    return new ArtistList(results);
  }

  /// Fetches a playlist by giving a [playlistID].
  ///
  /// See `https://docs-en.kkbox.codes/v1.1/reference#sharedplaylists-playlist_id`.
  @RequireAccessToken()
  Future<PlaylistInfo> fetchPlaylist(String playlistID,
      [Territory territory = Territory.taiwan]) async {
    final url = KKBOXOpenAPI._apiBase +
        'shared-playlists/$playlistID?territory=${_territoryToString(territory)}';
    final results = await this._get(url);
    return new PlaylistInfo(results);
  }

  /// Fetches tracks contained in a playlist by giving a [playlistID].
  ///
  /// See `https://docs-en.kkbox.codes/v1.1/reference#sharedplaylists-playlist_id-tracks`.
  @RequireAccessToken()
  Future<TrackList> fetchTracksInPlaylist(String playlistID,
      [Territory territory = Territory.taiwan,
      int offset = 0,
      int limit = 20]) async {
    final url = KKBOXOpenAPI._apiBase +
        'shared-playlists/$playlistID/tracks?territory=${_territoryToString(territory)}&offset=$offset&limit=$limit';
    final results = await this._get(url);
    return new TrackList(results);
  }

  /// Fetches featured playlists.
  ///
  /// See `https://docs-en.kkbox.codes/v1.1/reference#featuredplaylists`.
  @RequireAccessToken()
  Future<PlaylistList> fetchFeaturedPlaylists(
      [Territory territory = Territory.taiwan,
      int offset = 0,
      int limit = 100]) async {
    final url = KKBOXOpenAPI._apiBase +
        'featured-playlists?territory=${_territoryToString(territory)}&offset=$offset&limit=$limit';
    final results = await this._get(url);
    return new PlaylistList(results);
  }

  /// Fetches new hits playlists.
  ///
  /// See `https://docs-en.kkbox.codes/v1.1/reference#newhitsplaylists`.
  @RequireAccessToken()
  Future<PlaylistList> fetchNewHitsPlaylists(
      [Territory territory = Territory.taiwan,
      int offset = 0,
      int limit = 10]) async {
    final url = KKBOXOpenAPI._apiBase +
        'new-hits-playlists?territory=${_territoryToString(territory)}&offset=$offset&limit=$limit';
    final results = await this._get(url);
    return new PlaylistList(results);
  }

  /// Fetch featured playlists categories.
  ///
  /// See `https://docs-en.kkbox.codes/v1.1/reference#featuredplaylistcategories-category_id`.
  @RequireAccessToken()
  Future<FeaturedPlaylistCategoryList> fetchFeaturedPlaylistCategories(
      [Territory territory = Territory.taiwan,
      int offset = 0,
      int limit = 100]) async {
    final url = KKBOXOpenAPI._apiBase +
        'featured-playlist-categories?territory=${_territoryToString(territory)}&offset=$offset&limit=$limit';
    final results = await this._get(url);
    return new FeaturedPlaylistCategoryList(results);
  }

  /// Fetch featured playlists categories.
  ///
  /// See `(https://docs-en.kkbox.codes/v1.1/reference#featuredplaylistcategories-category_id-playlists`.
  @RequireAccessToken()
  Future<FeaturedPlaylistCategory> fetchFeaturedPlaylistsInCategory(
      String categoryID,
      [Territory territory = Territory.taiwan,
      int offset = 0,
      int limit = 100]) async {
    final url = KKBOXOpenAPI._apiBase +
        'featured-playlist-categories/$categoryID?territory=${_territoryToString(territory)}&offset=$offset&limit=$limit';
    final results = await this._get(url);
    return new FeaturedPlaylistCategory(results);
  }

  /// Fetch mood stations.
  ///
  /// See `https://docs-en.kkbox.codes/v1.1/reference#moodstations`.
  @RequireAccessToken()
  Future<RadioStationList> fetchMoodStations(
      [Territory territory = Territory.taiwan,
      int offset = 0,
      int limit = 100]) async {
    final url = KKBOXOpenAPI._apiBase +
        'mood-stations?territory=${_territoryToString(territory)}&offset=$offset&limit=$limit';
    final results = await this._get(url);
    return new RadioStationList(results);
  }

  /// Fetches tracks in a mood station.
  ///
  /// See `(https://docs-en.kkbox.codes/v1.1/reference#moodstations-station_id`.
  @RequireAccessToken()
  Future<RadioStation> fetchTracksInMoodStation(String stationID,
      [Territory territory = Territory.taiwan,
      int offset = 0,
      int limit = 100]) async {
    final url = KKBOXOpenAPI._apiBase +
        'mood-stations/$stationID?territory=${_territoryToString(territory)}&offset=$offset&limit=$limit';
    final results = await this._get(url);
    return new RadioStation(results);
  }

  /// Fetch genre stations.
  ///
  /// See `https://docs-en.kkbox.codes/v1.1/reference#genrestations`.
  @RequireAccessToken()
  Future<RadioStationList> fetchGenreStations(
      [Territory territory = Territory.taiwan,
      int offset = 0,
      int limit = 100]) async {
    final url = KKBOXOpenAPI._apiBase +
        'genre-stations?territory=${_territoryToString(territory)}&offset=$offset&limit=$limit';
    final results = await this._get(url);
    return new RadioStationList(results);
  }

  /// Fetches tracks in a genre station.
  ///
  /// See `https://docs-en.kkbox.codes/v1.1/reference#genrestations-station_id`.
  @RequireAccessToken()
  Future<RadioStation> fetchTracksInGenreStation(String stationID,
      [Territory territory = Territory.taiwan,
      int offset = 0,
      int limit = 100]) async {
    final url = KKBOXOpenAPI._apiBase +
        'genre-stations/$stationID?territory=${_territoryToString(territory)}&offset=$offset&limit=$limit';
    final results = await this._get(url);
    return new RadioStation(results);
  }

  /// Fetches new released album categories.
  ///
  /// See `https://docs-en.kkbox.codes/v1.1/reference#newreleasecategories`.
  Future<NewReleasedAlbumsCategoryList> fetchNewReleaseAlbumCategories(
      [Territory territory = Territory.taiwan,
      int offset = 0,
      int limit = 100]) async {
    final url = KKBOXOpenAPI._apiBase +
        'new-release-categories?territory=${_territoryToString(territory)}&offset=$offset&limit=$limit';
    final results = await this._get(url);
    return new NewReleasedAlbumsCategoryList(results);
  }

  /// Fetches a new released album category
  ///
  /// See `https://docs-en.kkbox.codes/v1.1/reference#newreleasecategories-category_id`.
  Future<NewReleasedAlbumsCategory> fetchNewReleasedAlbumInCategory(
      String categoryID,
      [Territory territory = Territory.taiwan,
      int offset = 0,
      int limit = 100]) async {
    final url = KKBOXOpenAPI._apiBase +
        'new-release-categories/$categoryID?territory=${_territoryToString(territory)}&offset=$offset&limit=$limit';
    final results = await this._get(url);
    return new NewReleasedAlbumsCategory(results);
  }

  /// Fetches charts.
  ///
  /// See `https://docs-en.kkbox.codes/v1.1/reference#charts`.
  Future<PlaylistList> fetchCharts(
      [Territory territory = Territory.taiwan,
      int offset = 0,
      int limit = 50]) async {
    final url = KKBOXOpenAPI._apiBase +
        'charts?territory=${_territoryToString(territory)}&offset=$offset&limit=$limit';
    final results = await this._get(url);
    return new PlaylistList(results);
  }

  /// Search in KKBOX.
  ///
  /// See `https://docs-en.kkbox.codes/v1.1/reference#search`.
  Future<SearchResults> search(String keyword,
      [List<SearchType> searchTypes = SearchType.values,
      Territory territory = Territory.taiwan,
      int offset = 0,
      int limit = 50]) async {
    var types = [];
    for (final aType in searchTypes) {
      switch (aType) {
        case SearchType.artist:
          types.add('artist');
          break;
        case SearchType.album:
          types.add('album');
          break;
        case SearchType.track:
          types.add('track');
          break;
        case SearchType.playlist:
          types.add('playlist');
          break;
      }
    }
    final typeString = types.join(',');

    final url = KKBOXOpenAPI._apiBase +
        'search?q=$keyword&type=$typeString&territory=${_territoryToString(territory)}&offset=$offset&limit=$limit';
    final results = await this._get(url);
    return new SearchResults(results);
  }
}
